
import os,string


def clean_scams(scam, extras={}):
    """
    clean_scams: Cleans all previous scams
    """
    try:
        print "  clean_scams: Cleaning all scams (howtos, macros...)"
        for i in scam.makeoptions:
            #print i
            #scam.addOption(i,'',False)
            scam.w.write("%s=\n"%i)
        scam.makeoptions={}
        scam.howtos=[]        # list of howtos the current scam contain
        scam.known_scamfiles={}  # files with rules known by the system, already parsed
        scam.performed_scams=[]  # scams that were performed correctly, do not rerun them
        scam.macro_helpers={}
        for i in filter(lambda x:x.endswith('_mhelper'),dir(scam)):
            Scam=scam.__class__
            f=eval('Scam.'+i)
            scam.macro_helpers.update({f.__doc__.split()[0].split('=')[0]+'=':f})
        scam.headers=[]
        scam.footers=[]
    except:
        import traceback
        traceback.print_exc()
    return True

# FIRST of all some basic scams: c and c++

def c(scam, extras={}):
    """
    c: Checks for the C compiler, and adds howtos. 
     [cc=cc]               Compiler to use
    """
    def howto_c(scam, target, src):
        if src[-2:]==".c":
            print "  c:Checking dependencies of "+src
            try:
                import os
                os.stat(src)
                p=os.popen("gcc -MM -MG "+string.join(scam.getOption("CXXFLAGS"))+" "+src)
                l=string.join(p.readlines()).split(":")
                scam.addRule(l[0],l[1:],[])
            except:
                scam.addRule(src[:-2]+".o",[src],[])
            return src[:-2]+".o"
        elif src[-2:]==".h":
            print "  c:Ignoring",src
            return ""
        return src

    defaults={'cc':'cc'}
    defaults.update(extras)
    
    cc=defaults['cc']
    bin=binary(cc)
    if bin=="":
        scam.error("%s not found"%cc)

    scam.addHowto(howto_c)
    scam.addOption("CC",cc,False)
    return True

cc=c


def cpp(scam, extras={}):
    """
    cpp: Checks for the C++ compiler, and adds howtos.
     [cc=g++,pch=no]              Compiler to use
    """
    def howto_cpp(scam, target, src):
        import os
        target=""
        if src.endswith(".cpp"):
          target=src[:-4]+".o"
          scam.cpp_files.append(src)
        elif src.endswith(".C"):
          target=src[:-2]+".o"
        elif src.endswith(".cc"):
          target=src[:-3]+".o"
        elif src[-2:]==".h":
            print "  cpp:Ignoring",src
            return ""
        if target!="":
            if not target in scam.rules:
                if scam.makeoptions['SCAM_CHECK_DEPENDS']==['full']:
                    print "  cpp:Checking dependencies of "+src
                    p=os.popen("gcc -MM -MG "+string.join(scam.getOption("CXXFLAGS"))+
                               " "+src+" 2>/dev/null")
                    rl=p.readlines()
                    l=string.join(rl).split(":")
                    scam.addRule(target,l[1:],[])
            return target
        return src

    def header_cpp(scam):
        scam.cpp_files=[]

    def file_exists(f):
        try:
            import os
            os.stat(f)
            return True
        except:
            return False

    def footer_cpp(scam):
        #print scam.cpp_files
        cpph=filter(lambda x:x.endswith('.cpp'),scam.cpp_files)
        cpph=filter(lambda x:file_exists(x),cpph)
        #print cpph

        if cpph:
            scam.addRule('.depends',[],
                         ['gcc -MM %s %s > .depends 2>/dev/null'%(string.join(scam.getOption("CXXFLAGS")),
                                                                              string.join(cpph))])
            scam.maketargets.append('.depends')
        else:
            scam.addRule('.depends',[],['touch .depends'])
        scam.clean+=['.depends']

        scam.w.write("include .depends")
        
    defaults={'cc':'g++', 'pch':'no', 'ld':-1, 'check_depends':'yes' }
    defaults.update(extras)
    
    cc=defaults['cc']
    bin=binary(cc)
    if bin=="":
        scam.error("%s not found"%cc)

    if defaults['pch']=='yes':
        scam.w.writelines("allheaders.h:\n"
                          "\tgrep -h \"#include <\" *.cpp *.h |sort|uniq > allheaders.h\n")
        scam.w.writelines("allheaders.h.gch: allheaders.h\n"
                          "\t$(CXX) $(CXXFLAGS) -c allheaders.h \n")
        scam.addOption("CXXFLAGS","-I.")
        scam.addOption("CXXFLAGS","-include allheaders.h")
        scam.addClean("allheaders.h.gch")
        scam.addClean("allheaders.h")
        scam.maketargets.append("allheaders.h.gch")

    scam.addOption("SCAM_CHECK_DEPENDS",defaults['check_depends'],False)
    if defaults['check_depends']=='yes':
        header_cpp(scam)
        scam.headers.append(header_cpp)
        scam.footers.append(footer_cpp)
    print '  cpp: check depends: %s'%defaults['check_depends']

    scam.addHowto(howto_cpp)
    scam.addOption("CXX",cc,False)
    if defaults['ld']!=-1:
        scam.addOption("CC",defaults['ld'],False)
    else:
        scam.addOption("CC",cc,False)
    return True


def ansi(scam,extras={}):
    """
    ansi: Adds ansi requirements to CFLAGS
    """
    scam.addOption("CFLAGS","-ansi");
    scam.addOption("CXXFLAGS","-ansi");
    return True

def expand(scam, extras={}):
    """
    expand: Expand wildcarded file names (*.cpp) to the real names that
         actually exist. if the there is no coicidence, returns the same
         name.
    """
    def howto_expand(scam, target, source):
        return scam.lookForFiles(source)
    
    scam.addHowto(howto_expand)
    return True

def backreference(scam, extras={}):
  """
    backreference: Looks for rules in the form {name} and looks back
                 for Make rules of that name. If any exist then the 
                 sources of such rules are appended to this one.
  """
  def howto_backreference(scam, target, source):
   try:
    if source[0]=='{' and source[-1]=='}':
      src=scam.getRule(source[1:-1])
      if len(src)==0:
        print "  backreference:%s reference do not exist or it is empty"%source
      else:
        print "  backreference:Expanded %s reference"%source
      scam.removeDist(source)
      return string.join(src)
   except:
    pass
   return source

  scam.addHowto(howto_backreference)
  return True
  
def staticlib(scam, extras={}):
    """
    staticlib: Adds the GNU Makefile rule to make static libraries
            from .o files.
    """
    scam.w.write("%.a:\n"+
                 "\trm -f $@\n"
                 "\tar rc $@ $^\n"
                 "\tranlib $@\n\n");
    return True

def sharedlib(scam, extras={}):
    """
    sharedlib: Adds the GNU Makefile rule to make shared libraries
            from .o files.
    """
    def sharedlib_writeRule(scam):
        scam.w.write("\n%.so:\n"+
                     "\trm -f $@\n"
                     "\tgcc -shared -Wl,-soname,$@ -o $@ $^ $(LDFLAGS)\n\n")
    sharedlib_writeRule(scam)
    scam.headers.append(sharedlib_writeRule)
    
    return True


def gl(scam,extras={}):
    """
    gl: Checks for OpenGL headers
    """
    Lib,lib=libs("GL")
    inc=include("gl")
    scam.addOption("CFLAGS","-I"+inc)
    scam.addOption("CXXFLAGS","-I"+inc)
    scam.addOption("LDFLAGS","-L"+Lib)
    scam.addOption("LDFLAGS","-l"+lib)
    return True

def jack(scam,extras={}):
    """
    jack: Checks for Jack library headers and libraries (http://jackit.sf.net)
    """
    Lib,lib=libs("jack")
    inc=include("jack/jack")
    scam.addOption("CFLAGS","-I"+inc)
    scam.addOption("CXXFLAGS","-I"+inc)
    scam.addOption("LDFLAGS","-L"+Lib)
    scam.addOption("LDFLAGS","-l"+lib)
    return True



def qt(scam,extras={}):
    """
    qt: Checks for QT environment. Also adds howtos.
    """
    def howto_qt(scam, target, source):
        if source[-3:]==".ui":
            base=source[:-3]
            scam.w.write("%s.h: %s\n"%(base,source))
            scam.w.write("\t$(UIC) -o %s.h %s\n\n"%(base,source))
            scam.addClean("%s.h"%base)
            scam.w.write("%s.cpp: %s %s.h\n"%(base,source,base))
            scam.w.write("\t$(UIC) -o %s.cpp -impl %s.h %s\n\n"%(base,base,source))
            scam.addClean("%s.cpp"%base)
            scam.w.write("%s.moc.cpp: %s.h\n\t$(MOC) -o %s.moc.cpp %s.h\n\n"%(
                base,base,base,base))
            scam.addClean("%s.moc.cpp"%base)
            return [base+".cpp", base+".moc.cpp"]
        
        elif source[-2:]==".h":
            fp=0
            qtheader=False
            try:
                fp=open(source)
                for i in fp.readlines():
                    if "Q_OBJECT" in i:
                        qtheader=True
                fp.close()
            except:
                scam.error("Cant open "+source)
            if qtheader:
                scam.w.write("%s.moc.cpp: %s\n\t$(MOC) -o %s.moc.cpp %s\n\n"%(
                    source[:-2],source,source[:-2],source))
                scam.addClean("%s.moc.cpp"%source[:-2])
                return source[:-2]+".moc.cpp"

        elif target.endswith(".ts"):
            scam.distfiles.append(target)
            
        return source

    scam.lookFor('cpp')

    import os
    qtdir=os.getenv("QTDIR","/usr/share/qt3")
    paths=[qtdir,"/usr/lib/qt3","/usr/","/usr/local"]

    incpath=map(lambda x:x+'/include',paths)
    inc=include("qt",incpath)
    if inc=="":
        scam.error("headers not found")
    scam.addOption("CXXFLAGS","-I"+inc)
    scam.addOption("CXXFLAGS","-DQT_THREAD_SUPPORT")
    print "  qt: Headers OK"

    libpath=map(lambda x:x+'/lib',paths)
    Lib,lib=libs("qt-mt",libpath)
    if lib=="":
        scam.error("libraries not found")
    scam.addOption("LDFLAGS","-L"+Lib)
    scam.addOption("LDFLAGS","-l"+lib)
    print "  qt: Libraries OK"

    for i in ['moc','uic','lupdate','lrelease']:
        bin=binary(i,map(lambda x:x+"/bin",paths))
        if bin=="":
            scam.error("%s not found"%i)
        else:
            scam.addOption("%s"%i.upper(),bin+"/"+i,False)

    scam.addRule("%.ts",[],
                 ["$(LUPDATE) `ls *.cpp 2>/dev/null` "+
                  "`ls *.h 2>/dev/null` $^ -ts $@"])
    scam.addRule("%.qm",["%.ts"],["$(LRELEASE) $^"])

    scam.addHowto(howto_qt)

    print "  qt: Ignoring %s"%str(extras)
    return True

def linguist(scam,extras={}):
    """
    linguist: Checks for linguist, and adds TS macro.
      [lupdate=lupdate]   Name for lupdate executable (linguist-qt4 for qt4)
      [lrelease=linguist] Name for lrelease executable (linguist-qt4 for qt4)
      [cpprootdir=..]     Directory where .cpp files are located (or common parent)
      [sourcefileextension=cpp] Extension for source files to search for.
      
      Also important are MACROS: TS_INSTALLDIR=where to install ts files (else not installed)
                               TS=list of ts files
      It is important to set TS_INSTALLDIR before TS.
    """
    
    def TS_mhelper(scam,l):
            """
            Helper for TS files. Just add ts files and it creates qm rules.
            """
            l=l[3:].split()
            qms=map(lambda x:x.replace('.ts','.qm'),l)
            scam.addRule('TS',qms,[])
            scam.maketargets.append('TS')
            for i in l:
                    qm=i.replace('.ts','.qm')
                    scam.addRule(qm,[i],[])
            
            print ' * Added TS rules for: %s'%string.join(qms)
            if 'TS_INSTALLDIR' in scam.makeoptions:
                    print ' * Install TS to: %s'%scam.getOption('TS_INSTALLDIR')[0]
                    scam.addInstall('$(TS_INSTALLDIR)'+':['+string.join(qms,',')+']')
            
    
    defaults={'lupdate':'lupdate','lrelease':'lrelease','cpprootdir':'..','sourcefileextension':'cpp'}
    defaults.update(extras)
    lupdate_bin=binary(defaults['lupdate'])+'/'+defaults['lupdate']
    lrelease_bin=binary(defaults['lrelease'])+'/'+defaults['lrelease']
    
    scam.addRule('%.qm',['%.ts'],['%s $<'%lrelease_bin])
    scam.addRule('%.ts',['remake-tsfiles'],['%s `find %s -name "*.%s"` -ts $@'%
                                              (lupdate_bin,defaults['cpprootdir'],defaults['sourcefileextension'])])
    scam.addRule('remake-tsfiles',[],['touch $@'])
    scam.addClean('remake-tsfiles')
    scam.macro_helpers.update({'TS=':TS_mhelper})
    return True

def qt4(scam,extras={}):
    """
    qt4: Checks for QT environment. Also adds howtos.
      [qtlibs=QtCore QtGui] to add new libs
    """
    def howto_qt4(scam, target, source):
        if source.endswith(".ui"):
            scam.addDist(source)
            base=source[:-3]
            scam.addRule("ui_%s.h"%base,[source],["$(UIC) -o ui_%s.h %s"%(base,source)])
            scam.addClean("ui_%s.h"%base)
            
            scam.addRule("%s.moc.cpp"%base,["%s.h"%base],
                                   ["$(MOC) -o %s.moc.cpp %s.h"%(base,base)])
            scam.addClean("%s.moc.cpp"%base)
            print "  qt4:Rule for ui file %s"%source
            return [base+".cpp", base+".moc.cpp"]
            
        elif source.endswith(".h"):
            fp=0
            qtheader=False
            try:
                fp=open(source)
                for i in fp.readlines():
                    if "Q_OBJECT" in i:
                        qtheader=True
                fp.close()
            except:
                scam.error("Cant open "+source)
            if qtheader:
                base=source[:-2]
                scam.addRule("%s.moc.cpp"%base,["%s.h"%base],
                                        ["$(MOC) -o %s.moc.cpp %s.h"%(base,base)])
                scam.addClean("%s.moc.cpp"%source[:-2])
                print "  qt4:Rule to compile moc for Q_OBJECT at file %s"%source
                return source[:-2]+".moc.cpp"
        return source

    myextras={"qtlibs":"QtCore QtGui","extra":""}
    myextras.update(extras)

    import os
    qtdir=os.getenv("QTDIR","/usr/share/qt4");
    paths=["/usr/share/qt4",qtdir,"/usr/lib/qt4","/usr/"]
    components=["QObject","QWidget","QtCore/QObject"]
    components+=string.split(myextras['extra'])
    
    incpath=map(lambda x:x+'/include',paths)
    for i in components:
        inc=include(i,incpath,"")
        if inc=="":
            scam.error("headers not found")
        scam.addOption("CXXFLAGS","-I"+inc)
    scam.addOption("CXXFLAGS","-DQT_THREAD_SUPPORT")
    print "  qt4: Headers OK"

    libpath=map(lambda x:x+'/lib',paths)
    libpath=["/usr/lib/x86_64-linux-gnu"] + libpath
    for i in myextras["qtlibs"].split(" "):
        Lib,lib=libs(i,libpath)
        if lib=="":
            scam.error("libraries not found")
        scam.addOption("LDFLAGS","-L"+Lib)
        scam.addOption("LDFLAGS","-l"+lib)
    print "  qt4: Libraries OK"

    bin=binary("moc",[qtdir+"/bin","/usr/bin/","/usr/local/bin"])
    if bin=="":
        scam.error("moc not found")
    scam.addOption("MOC",bin+"/moc",False)
    bin=binary("uic",[qtdir+"/bin","/usr/bin/","/usr/local/bin"]) 
    if bin=="":
        scam.error("uic not found")
    scam.addOption("UIC",bin+"/uic",False)
    print "  qt4: Binaries OK"

    scam.addHowto(howto_qt4)

    print "  qt4: Adding libs %s"%str(myextras["qtlibs"])
    return True

def kde(scam, extras={}):
    """
    kde: Looks for the KDE environment.
    """
    
    inc=include("kapp",["/usr/include/kde","/usr/include"])
    if inc=="":
        scam.error("headers not found")
    Lib,lib=libs("kdecore")
    if lib=="":
        scam.error("libraries not found")
    Lib,lib=libs("kdeui")
    if lib=="":
        scam.error("libraries not found")
    Lib,lib=libs("kio")
    if lib=="":
        scam.error("libraries not found")
    scam.addOption("CXXFLAGS","-I"+inc)
    scam.addOption("LDFLAGS","-L"+Lib)
    scam.addOption("LDFLAGS","-l"+lib)

    if extras!="":
        print "  kde: Ignoring %s"%str(extras)
    
def lexyacc(scam, extras={}):
    """
    lexyacc, yacclex, lex, yacc: Look for lex and yacc environments.
    """
    def howto_lex(scam, target,src):
        if src[-4:]==".lex":
            target=src[:-4]+".c"
            scam.w.writelines("%s: %s\n"%(target,src))
            scam.w.writelines("\t$(LEX) $(LEXFLAGS) -o%s %s\n\n"%(target,src))
            return target
        return src

    def howto_yacc(scam, target,src):
        if src[-2:]==".y":
            target=src[:-2]+".c"
            targeth=src[:-2]+".h"
            scam.w.writelines("%s: %s\n"%(target,src))
            scam.w.writelines("\t$(YACC) -d -o %s %s $(YACCFLAGS)\n\n"%(target,src))
            scam.w.writelines("%s: %s\n"%(targeth,src))
            scam.w.writelines("\t$(YACC) -d -o %s %s $(YACCFLAGS)\n\n"%(target,src))
            return target
        return src
    binl=binary("lex")
    if binl=="":
        scam.error("lex not found")
    biny=binary("yacc")
    if biny=="":
        scam.error("yacc not found")
    scam.addOption("LEX",binl+"/lex",False)
    scam.addOption("YACC",biny+"/yacc",False)

    
    scam.addHowto(howto_lex)
    scam.addHowto(howto_yacc)
    return True

yacclex=lexyacc
lex=yacclex
yacc=lexyacc


def doxygen(scam, extras={}):
    """
    doxygen: Adds the doc: target that autogenerates source documentation with Doxygen.
    """
    bin=binary("doxygen")
    scam.addOption("DOXYGEN",bin+"/doxygen",False);
    scam.w.writelines("doc:\n")
    scam.w.writelines("\t$(DOXYGEN)\n")
    return True



def alsa(scam,extras={}):
    """
    alsa: Checks for ALSA libraries and headers
    """
    lib=libs("asound")
    inc=include("asoundlib.h")
    if inc!="":
        scam.addOption("CFLAGS","-I"+inc)
        scam.addOption("CXXFLAGS","-I"+inc)
    scam.addOption("LDFLAGS","-L"+lib[0])
    scam.addOption("LDFLAGS","-l"+lib[1])
    return True

def gstreamer(scam,extras={}):
    """
    gstreamer: Looks for gstreamer libraries and headers
     [version=0.8]           Version to look for 
    """
    scam.lookFor('glib[version=2.0]')

    opts={'version':'0.8'}
    opts.update(extras)
    lib=libs("gstreamer-%s"%opts['version'])
    inc=include("gst",["/usr/include/gstreamer-%s/gst/"%opts['version']])
    if inc=="":
        scam.error("gst.h not found")  
    scam.addOption("CFLAGS","-I"+inc[:-4])
    scam.addOption("CXXFLAGS","-I"+inc[:-4])
    scam.addOption("LDFLAGS","-L"+lib[0])
    scam.addOption("LDFLAGS","-l"+lib[1])
    return True

                
def glib(scam, extras={}):
    """
    glib: Looks for glib libraries and headers
     [version=2.0]           Version to look for 
    """
    scam.lookFor('libxml2')

    opts={'version':'2.0'}
    opts.update(extras)

    lib=libs("glib-%s"%opts['version'])
    scam.addOption("LDFLAGS","-L"+lib[0])
    scam.addOption("LDFLAGS","-l"+lib[1])

    inc=include("glib",["/usr/include/glib-%s/"%opts['version']])
    if inc=="":
        scam.error("glib.h not found")  
    scam.addOption("CFLAGS","-I"+inc)
    scam.addOption("CXXFLAGS","-I"+inc)

    inc=include("glibconfig",["/usr/lib/glib-%s/include/"%opts['version']])
    if inc=="":
        scam.error("glibconfig.h not found")  
    scam.addOption("CFLAGS","-I"+inc)
    scam.addOption("CXXFLAGS","-I"+inc)
    return True


def libxml2(scam, extras={}):
    """
    libxml2: Looks for libxml2 library and headers
    """

    lib=libs("xml2")
    scam.addOption("LDFLAGS","-L"+lib[0])
    scam.addOption("LDFLAGS","-l"+lib[1])

    inc=include("libxml/parser")
    if inc=="":
        scam.error("libxml/parser.h not found")  
    scam.addOption("CFLAGS","-I"+inc)
    scam.addOption("CXXFLAGS","-I"+inc)
    return True

def pythonc(scam, extras={}):
    """
    pythonc: Looks fot the python C development stuff for extending
    """
    inc=include("Python")
    if inc=="":
        scam.error("Python.h not found")  
    scam.addOption("CFLAGS","-I"+inc)
    scam.addOption("CXXFLAGS","-I"+inc)
    return True

    
